/**
 * Stars!
 * 思考ルーチンクラス
 *
 * $Id: stars_ai.js,v 1.11 2006/02/13 14:37:12 shio Exp $
 */

//思考ルーチン
function Ai(app, empire) {
    this.app = app;
    this.empire = empire;
    this.tecNavigate = new TecNavigate(this.empire);
    this.tecSearch = new TecSearch(this.empire);
    this.tecMilitary = new TecMilitary(this.empire);

    // 惑星ごとの行動を考える
    this.thinkPlanets = function() {
       //var pList = this.empire.makePlanetList(app.starList);
       var pList = this.empire.ctlPlanetList;
       pList = pList.randomize();
        for (var i = 0; i < pList.length; i++) {
            var p = pList[i];

            this.countUchuju = app.uchuju.countContainsUchuju(p.loc, this.empire.fnd);
            var cmdList = this.getCmdList(this.empire, p.loc);
            for (var j = 0; j < cmdList.length; j++) {
                var cmd = cmdList[j];
                if (cmd.think(this, p) ) {
                    this.empire.tmp_trs -= cmd.cost();
                    app.commandList.push(cmd);
                    break;
                }
            }
        }
    }

    // 防衛力強化が必要かどうかを判定する
    this.checkBuildDfc = function(p) {
        if (this.empire.tmp_trs < BASE_COST) {
              return false;
        }
        if ( (p.dfc + 1)  > parseInt(p.dpl / 2 + 0.5) ) {
              return false;
        }
        return true;
    }

    // 艦隊の建設が必要かどうかを判定する
    this.checkBuildFleet = function(p) {
        if (this.empire.tmp_trs < FLEET_BUILD_COST) {
            return false;
        }
        if (p.ecn < 5) {
            return false;
}
        if (this.countUchuju > this.empire.countFleet() ) {
            if (this.empire.countFleet() >= (this.empire.fleetNeeds() * 2) ) {
                return false;
            }
        } else if (this.empire.countFleet() >= this.empire.fleetNeeds() ) {
            return false;
        }

        return true;
    }

    // 植民地の建設が必要かどうかを判定する
    this.checkBuildColony = function(p) {
        if (app.uchuju.nvg > 1) {
            if (this.countUchuju > 0) {
                return false;
            }
        }
        if (this.empire.tmp_trs < 50) {
            return false;
        }
        if (p.dpl < 6) {
            return false;
        }
        return true;
    }

    // 移民船団の建設が必要かどうかを判定する
    this.checkBuildSettler = function(p) {
        if (this.empire.tmp_trs < SETTLER_BUILD_COST) {
            return false;
        }
        if (p.dpl < 8) {
            return false;
        }

        // 到達可能な未開拓惑星が発見されている->false
        var undplList = this.findUndplList();
        for (var lk in undplList) {
            if (this.checkBetweenNavigatable(lk, p.loc, this.empire.nvg,
                    this.empire.ctlPlanetList)) {
                return false;
            }
        }
        
        // 人口が３以下で到達可能な植民星が存在している->false
        for (var i = 0; i < this.empire.ctlPlanetList.length; i++) {
            var p2 = this.empire.ctlPlanetList[i];
            if (p2.dpl > 3) {
                continue;
            }
            if (this.checkBetweenNavigatable(p2.loc, p.loc, this.empire.nvg,
                    this.empire.ctlPlanetList)) {
                return false;
            }
        }

        // ここまできたら、移民船団を建設する。
        return true;
    }

    // 技術の進展を考える
     this.thinkTec = function() {
        var dpl_p = 0;
        var fnd_p = 0;
        var nvg_p = 0;
        var mlt_p = 0;

        var e = this.empire;
        var totalDpl = e.totalDpl();

        // 人口に比べ経済力が低い場合には開発に肩入れする
        if (totalDpl > (e.totalEcn() * 0.7) ) {
            dpl_p += 3;
        }
        if (totalDpl > e.totalEcn() ) {
            dpl_p += 3;
        }
        if (totalDpl > e.totalEcn() * 2 ) {
            dpl_p += 3;
        }

        // 経済力が最大経済力の８割以下ならば、開発にてこ入れする。
        if ( this.empire.totalEcn() <= (this.empire.calTotalMaxEcn() * 0.8)) {
            dpl_p += 3;
        }
        
        // 十分に資金があれば、軍事に肩入れする。
        if (e.trs > 100) {
            mlt_p = 1;
        } else {
            dpl_p += 3;
        }

        // 植民可能な星がみつかっていない場合には、探査を優先する
        var undplList = this.findUndplList();
        var nearest = this.findNearestUndpl(undplList);

        if (! nearest) {
            fnd_p++;
        // 既に植民可能な星がみつかっているが、航力が不足する場合には、航宙力を優先する
        } else if (nearest.distance > e.nvg) {
            nvg_p++;
        }

        // 帝国内に飛び地がある場合には、航宙力を優先する
        if (! e.isReachAllPlanets()) {
            nvg_p++;
        }

        // 貿易相手に比べて負けている技術に肩入れする
        if (app.trade.tDic[e.id].max_fnd) {
            fnd_p++;
        }
        if (app.trade.tDic[e.id].max_nvg) {
            nvg_p++;
        }
        if (app.trade.tDic[e.id].max_mlt) {
            mlt_p += 2;
        }

        // 予算配分を決定する。
        var total_p = dpl_p + fnd_p + nvg_p + mlt_p;
        if (total_p > 0) {
            this.tecSearch.yosan_p = parseInt(fnd_p * 10 / total_p) * 10;
            this.tecNavigate.yosan_p = parseInt(nvg_p * 10 / total_p) * 10;
            this.tecMilitary.yosan_p = parseInt(mlt_p * 10 / total_p) * 10;
        } else {
            this.tecSearch.yosan_p = 0;
            this.tecNavigate.yosan_p = 0;
        }
    }

    this.developTec = function() {
        // 探査力を強化
        this.tecSearch.doDevelop();
        // 航宙力を強化
        this.tecNavigate.doDevelop();
        // 軍事技術を強化
        this.tecMilitary.doDevelop();
        // 遭遇チェック
        this.findUndplList();
    }

    // 移民船団の行動を考える
    this.thinkSettlers = function() {
        var sList = this.empire.settlerList;
        sList = sList.randomize();
        for (var i = 0; i < sList.length; i++) {
            var s = sList[i];

            var cmdList = this.getSettlerCmdList(this.empire, s);
            for (var j = 0; j < cmdList.length; j++) {
                var cmd = cmdList[j];
                if (cmd.think(this) ) {
                    this.empire.tmp_trs -= cmd.cost();
                    app.settlerCmdList.push(cmd);
                    break;
                }
            }
        }
    }

    // 最も近い未開拓の惑星を返す { loc, distance }
    this.findNearestUndpl = function(undplList) {
        var min = 9999;
        var minLk = null;
        var lkList = new Array();
        for (var lk in undplList) {
            lkList.push(lk);
        }
        lkList = lkList.randomize();
        for (var i = 0; i < lkList.length; i++) {
            var lk = lkList[i];
            if (undplList[lk]+0 < min) {
                min = undplList[lk];
                minLk = lk;
            }
        }
        if (minLk) {
            return { loc: minLk, distance: min };
        }
        return null;
    }

    // 植民可能な未開拓惑星を返す
    this.findColonizableList = function(undplList, nvg) {
        var result = new Array();
        var empire = this.empire;
        for (var lk in undplList) {
            var d = parseInt(undplList[lk]);
            if (d <= nvg) {
                result.push({ loc: lk, distance: undplList[lk] });
            }
        }
        return result;
    }

    // 未開拓の惑星を探す -> key:loc, value=distanceの連想配列
    this.findUndplList = function() {
        var e = this.empire;
        var pList = e.ctlPlanetList;
        var fList = new Object();
        var list = app.starList;

        for (var lk in list) {
            if (! e.visibleList[lk]) {
                continue;
            }
            var p2 = app.starList[lk];
            var e2 = p2.ctl;
            if (p2.ctl == null) {
                fList[lk] = e.getNearestPlanetDistance(lk);
            } else if (e2.id != e.id) {
                if (e.yoronList[e2.id]) {
                    continue;
                }
                // 航行可能判定
                var d = e.getNearestPlanetDistance(lk);
                if (d <= e.nvg) {
                    e.yoronList[e2.id] = 50;  // 初期値
                    e2.yoronList[e.id] = 50;
                    printConsole(e.id, e.name + "は"
                            + p2.name + "で" + p2.ctl.name + "に遭遇");
                    break;
                }
            }
            if (e.foundPlanetList.contains(lk) ) {
            } else {
                if (p2) {
                    e.foundPlanetList.push(lk);
                    printConsole(e.id, e.name + "は" + lk + "に惑星" + p2.id + "を発見");
                }
            }
        }
        
        return fList;
    }
    // 指定された惑星から、未知の惑星を探査する
    this.findUndplListFromLoc = function(loc, fnd) {
        var result = new Object();
        for (var lk in app.starList) {
            var d = calDistance(loc, lk);
            if (! this.empire.visibleList[lk]) {
                continue;
            }
            var p = app.starList[lk];
            if (p == null) {
                continue;
            }
            if (p.ctl == null) {
                result[lk] = d;
            }
        }

        return result;
    }

    // 指定された惑星から、未知の惑星を探査する
    this.findPlanetListFromList = function(loc, fnd, list) {
        var result = new Object();
        for (var lk in list) {
            var p = list[lk];
            var d = calDistance(loc, lk);
            if (d > fnd) {
                continue;
            }
            result[lk] = d;
        }

        return result;
    }

    // 指定された範囲内にある別の惑星を探す
    this.findPlanetList = function(loc, distance1, findList, distance2) {
        var fList = (findList) ? findList : new Object();
        var d2 = (distance2) ? distance2 : 0;
        var chkList = new Object();
        var list = null;
        var lk = null;
        // 隣を探す
        if (distance1 >= 2) {
            list = app.ajustList(loc);
            for (var i = 0; i < list.length; i++) {
                lk = list[i];
                chkList[lk] = d2 + 2;
            }
        }
        if (distance1 >= 3) {
            list = app.obliqueList(loc);
            for (var i = 0; i < list.length; i++) {
                lk = list[i];
                chkList[lk] = d2 + 3;
            }
        }
        fList[loc] = distance2;
        for (lk in chkList) {
            if (lk in fList) {
            } else {
                fList[lk] = chkList[lk];
                fList = this.findPlanetList(
                        lk, distance1 - chkList[lk], fList, chkList[lk]);
            }
        }
        var result = new Object;
        for (lk in fList) {
            if (lk in app.starList) {
                result[lk] = fList[lk];
            }
        }

        return result;
    }

    /**
     * コマンドリストを返す。
    */
    this.getCmdList = function(e, ploc) {
        var p = app.starList[ploc];

        var result = new Array();
        var cmd = new CmdBuildColony(e, p);
        result.push(cmd);

        cmd = new CmdBuildDfc(e, p);
        result.push(cmd);

        cmd = new CmdBuildEcn(e, p);
        result.push(cmd);

        cmd = new CmdBuildFleet(e, p);
        result.push(cmd);

        cmd = new CmdImin(e, p);
        result.push(cmd);

        cmd = new CmdBuildSettler(e, p);
        result.push(cmd);

        return result;
    }

    /**
     * 植民船団コマンドリストを返す。
    */
    this.getSettlerCmdList = function(e, loc) {
        var result = new Array();
        
        result.push(new CmdSettlerColony(e, loc));
        
        result.push(new CmdSettlerJoin(e, loc));
        
        result.push(new CmdSettlerMove(e, loc));

        return result;
    }

    // 航行で到達可能かどうかを調べる
    this.checkBetweenNavigatable = function(sloc, eloc, nvg, starList) {
        var starHash = new Object();
        for (var i = 0; i < starList.length; i++) {
            starHash[starList[i].loc] = false;
        }

        var clist = new Array();
        clist.push(sloc);
        if (starHash[sloc] == 0) {
            starHash[sloc] = true;
        }
        while(clist.length > 0) {
            var nlist = new Array();
            for (var i = 0; i < clist.length; i++) {
                var cloc = clist[i];
                if (calDistance(cloc, eloc) <= nvg) {
                    return true;
                }
                for (var lk in starHash) {
                    if (starHash[lk]) {
                        continue;
                    }
                    if (calDistance(cloc, lk) <= nvg) {
                        starHash[lk] = true;
                        nlist.push(lk);
                    }
                }
            }
            clist = nlist;
        }
        return false;
    }
}
