/**
 * Stars!
 * 宇宙怪獣クラス群
 *
 * $Id: stars_uchuju.js,v 1.9 2006/02/12 11:07:30 shio Exp $
 */

// 宇宙怪獣モジュール
function Uchuju(ritsu) {
    this.uchujuList = new Array();
    this.ritsu = ritsu;
    this.uid = 0;

    /**
      * 宇宙怪獣存在チェック.
      * @param loc 位置情報(r1c1形式)
      * @return 宇宙怪獣部隊 or null
      */
    this.isUchujuLoc = function(loc) {
        for (var i = 0; i < this.uchujuList.length; i++) {
            var u = this.uchujuList[i];
            if (u && (loc == u.loc) && (u.force > 0) ) {
                return u;
            }
        }

        return null;
    }

    /**
     * 宇宙怪獣を発生させる
     */
    this.createUchuju = function() {
        var d = app.rollDice(100);
        if (d <= this.ritsu) {
            // 宇宙怪獣発生
            this.calTec();
            var loc = "r" + (app.rollDice(app.height) - 1)
                    + "c" + (app.rollDice(app.width) - 1);
            var u = new UchujuUnit(this, this.uid, loc, this.force);
            this.uid++;
            this.uchujuList.push(u);
            printConsoleByLoc(loc, loc + "で宇宙怪獣が発生");
        }
    }

    /**
     * 宇宙怪獣を取り除く
     * @param uid   宇宙怪獣ID
     */
    this.removeUchuju = function(uid) {
        for (var i = 0; this.uchujuList.length; i++) {
            var u = this.uchujuList[i];
            if (uid == u.id) {
                this.uchujuList[i] = null;
                this.uchujuList = this.uchujuList.remove(i);
                return;
            }
        }
    }

    /**
     * すべての宇宙怪獣を行動させる
     */
    this.doActionAll = function() {
        var battle = new Battle();
        for (var i = 0; i < this.uchujuList.length; i++) {
            var u = this.uchujuList[i];
            if (u == null) {
                continue;
            }
            if (u.force == 0) {
                continue;
            }
            u.doAction();
            u.doBattle(battle);
        }
    }

    /**
     * 宇宙怪獣の技術レベル／戦力を決定する
     */
    this.calTec= function() {
        var fnd = 0;
        var nvg = 0;
        var mlt = 0;
        var fleet = 0;
        var countEmp = app.empireList.length;
        for (var eid = 0; eid < app.empireList.length; eid++) {
            var e = app.empireList[eid];
            fnd += e.fnd;
            nvg += e.nvg;
            mlt += e.mlt;
            fleet += e.countFleet();
        }
        this.fnd = parseInt(fnd / countEmp) + 2;
        this.nvg = parseInt(nvg / countEmp) - 1;
        this.mlt = parseInt(mlt / countEmp) - 1;
        this.force = parseInt(fleet / countEmp + 0.5) + 1;
    }

    this.countContainsUchuju = function(loc, dis) {
        var result = 0;
        for (var i = 0; i < this.uchujuList.length; i++) {
            var unit = this.uchujuList[i];
            if ( (! unit) || (unit.force == 0) ) {
                continue;
            }
            if (calDistance(loc, unit.loc) <= dis) {
                result += unit.force;
            }
        }

        return result;
    }

    this.clearZero = function() {
        var newList = new Array();
        for (var i = 0; i < this.uchujuList.length; i++) {
            var unit = this.uchujuList[i];
            if ( (! unit) || (unit.force == 0) ) {
                continue;
            }
            newList.push(unit);
        }
        this.uchujuList = newList;
    }
}

/**
 * 宇宙怪獣部隊クラス
 * @param uid   宇宙怪獣ID
 * @param loc   位置
 * @param fnd   索敵範囲
 * @param nvg   移動範囲
 * @param mlt   軍事レベル
 * @param force 戦力
 */
function UchujuUnit(mgr, uid, loc, force) {
    this.mgr = mgr;
    this.uid = uid;
    this.loc = loc;
    this.force = force;

    /**
     * 地図上に表示するポップアップテキストを返す.
     */
    this.getTitle = function() {
        return ("探-" + mgr.fnd +" 航-" + mgr.nvg + " 軍-" + mgr.mlt
                + " 数-" + this.force);
    }

    /**
     * 宇宙怪獣が移動を行なう.
     */
    this.doAction = function() {
        // 移動に十分な観測技術や航行技術を有していない場合は移動しない
        if ( (this.mgr.fnd <= 1) || (this.mgr.nvg <= 1) ) {
            return;
        }

        // 既に移民船団と同じ場所にいるなら移動しない
        if (app.settlerMap[this.loc] >= 0) {
            return;
        }
        // すでに植民済の惑星に到達している時には移動しない.
        var planet = app.starList[this.loc];
        if (planet && planet.ctl) {
            return;
        }
        
        // 移民船団を探して移動する。
        var nearestD = 9999;
        var nvgList = new Array();
        var nearestList = new Array();
        for (var eix = 0; eix < app.empireList.length; eix++) {
            var e = app.empireList[eix];
            for (var six = 0; six < e.settlerList.length; six++) {
                var lk = e.settlerList[six];
                var d = calDistance(lk, this.loc);
                if (d <= this.mgr.nvg) {
                    nvgList.push(lk);
                    continue;
                }
                if (nvgList.length > 0) {
                    continue;
                }
                if (d == nearestD) {
                    nearestList.push(lk);
                } else if (d < nearestD) {
                    nearestList = new Array();
                    nearestList.push(lk);
                    nearestD = d;
                }
            }
        }

        if (nvgList.length == 0) {
            // 近隣の惑星を探して移動する.
            for (var lk in app.starList) {
                planet = app.starList[lk];
                if (! planet.ctl) {
                    continue;
                }
                var d = calDistance(lk, this.loc);
                if (d > this.mgr.fnd) {
                    continue;
                }
                if (d <= this.mgr.nvg) {
                    nvgList.push(lk);
                    continue;
                }
                if (nvgList.length > 0) {
                    continue;
                }
                if (d == nearestD) {
                    nearestList.push(lk);
                } else if (d < nearestD) {
                    nearestList = new Array();
                    nearestList.push(lk);
                    nearestD = d;
                }
            }
        }
        
        if (nvgList.length > 0) {
            lk = this.choiceWeakLoc(nvgList);
            this.moveTo(lk);
        } else if (nearestList.length > 0) {
            lk = this.choiceWeakLoc(nearestList);
            this.moveTo(planet.loc);
        }
    }

    /**
     * 一番弱い星を攻撃目標とする.
     */
    this.choiceWeakLoc = function(list) {
        list = list.randomize();
        var minForce = 9999;
        var minLevel = 9999;
        var minLoc = null;
        for (var i = 0; i < list.length; i++) {
            // 移民船団を最優先に目標とする。
            var loc = list[i];
            if (app.settlerMap[loc] >= 0) {
                minLoc = loc;
                break;
            }
            // 惑星の目標判定
            var planet = app.starList[loc];
            var force = planet.dfc;
            if (force < minForce) {
                minForce = force;
                minLoc = loc;
                minLevel = planet.ctl.mlt;
            } else if (force == minForce) {
                if (planet.ctl.mlt < minLevel) {
                    minLoc = loc;
                    minLevel = planet.ctl.mlt;
                }
            }
        }
        return minLoc;
    }

    /**
     * 目標地点に向かって移動する。
     */
    this.moveTo = function(lk) {
        this.loc = moveTo(this.loc, lk, this.mgr.nvg);
        printConsoleByLoc(this.loc, "宇宙怪獣が" + this.loc + "に移動");
    }

    this.makeForceList = function() {
        var list = new Array();
        for (var i = 0; i < this.force; i++) {
            list.push(new Force(this.mgr.nvg, this.mgr.mlt, this));
        }

        return list;
    }

    this.doBattle = function(battle) {
        var planet = app.starList[this.loc];
        var empire = app.findSettlerEmpire(this.loc);
        var isSettler = (empire) ? true : false;
        if ( (planet) && (! empire) ) {
            empire = planet.ctl;
        }
        
        if (empire) {
            var locname = (planet) ? planet.name : "移民船団(" + this.loc + ")近辺";
            var msgid = empire.id;
            printConsole(msgid, "宇宙怪獣が" + locname + "を攻撃！");
            var fleet = empire.getReachedFleetList(this.loc);
            if (fleet.length > 0) {
                printConsole(msgid, locname + "で" + empire.name
                        + "艦隊("  + fleet.length
                                + ")と宇宙怪獣(" + this.force + ")の戦闘が発生！");
            }
            var j = 0;
            for (j = 0; j < 3; j++) {
                if ( (fleet.length <= 0) || (this.force <= 0)) {
                    break;
                }
                printConsole(msgid, "第" + (j+1) + "ラウンド");
                var ufleet = this.makeForceList();
                var efleet = battle.getFleetForceList(fleet)
                battle.conbatBattle(efleet, ufleet);
                battle.updateExpFleet(efleet);
                if (battle.damageList1.length > 0) {
                    fleet = battle.applyBattleEmpireFleet(empire, fleet, battle.damageList1);
                }
                if (battle.damageList2.length > 0) {
                    battle.applyBattleUchuju(msgid, this, battle.damageList2);
                }
            }
            if (fleet.length > 0) {
                return;
            }
            if (this.force <= 0) {
                return;
            }

            if (isSettler) {
                printConsole(msgid, locname + "で" + "宇宙怪獣(" + this.force
                        + ")が" + empire.name + "の植民船団を攻撃！");
                for ( ; j < 3; j++) {
                    printConsole(msgid, "第" + (j+1) + "ラウンド");
                    var fleet = this.makeForceList();
                    var sfleet = battle.makeSettlerForceList(this.loc);
                    battle.conbatBattle(sfleet, fleet);
                    if (battle.damageList1.length > 0) {
                        battle.applyBattleSettler(empire, this.loc);
                    }
                    if (battle.damageList2.length > 0) {
                        battle.applyBattleUchuju(msgid, this, battle.damageList2);
                    }
                    if (battle.damageList1.length > 0) {
                        break;
                    }
                }
            }
            if (this.force <= 0) {
                return;
            }

            if (! planet) {
                return;
            }
            printConsole(msgid, "宇宙怪獣(" + this.force
                    + ")が惑星" + planet.name + "守備隊("
                            + planet.dfc + ")を攻撃！");
            for ( ; j < 3; j++) {
                if ((this.force <= 0) || (planet.dfc <= 0) ) {
                    break;
                }
                printConsole(msgid, "第" + (j+1) + "ラウンド");
                if (planet.dfc > 0) {
                    var fleet = this.makeForceList();
                    var pfleet = planet.makeForceList();
                    battle.conbatBattle(pfleet, fleet);
                    if (battle.damageList1.length > 0) {
                        battle.applyBattlePlanetDfc(planet, battle.damageList1);
                    }
                    if (battle.damageList2.length > 0) {
                        battle.applyBattleUchuju(msgid, this, battle.damageList2);
                    }
                }
            }
            
            // 惑星戦闘
            if ( (this.force > 0) && (planet.dfc <= 0) ) {
                printConsole(msgid, "宇宙怪獣(" + this.force
                        + ")が惑星" + planet.name + "市民("
                                + planet.dpl + ")を攻撃！");
                for ( ; j < 3; j++) {
                    if ((this.force <= 0) || (planet.dpl <= 0) ) {
                        break;
                        printConsole(msgid, "第" + (j+1) + "ラウンド");
                    }
                    if (planet.dpl > 0) {
                        var fleet = this.makeForceList();
                        var pfleet = planet.makeDplForceList();
                        battle.conbatBattle(pfleet, fleet);
                        if (battle.damageList1.length > 0) {
                            battle.applyBattlePlanetDpl(planet, battle.damageList1);
                        }
                        if (battle.damageList2.length > 0) {
                            battle.applyBattleUchuju(msgid, this, battle.damageList2);
                        }
                    }
                }
            }
        }
    }
}